# Copyright 2024 The bayesnf Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model code for Bayesian-neural-field time-series work."""

import enum
from typing import Tuple

import flax
from flax import linen as nn
import jax
from jax import numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp  # pylint: disable=g-importing-member

tfd = tfp.distributions


class LikelihoodDist(enum.Enum):
  NORMAL = 'NORMAL'
  NB = 'NB'
  ZINB = 'ZINB'


def make_seasonal_frequencies(
    seasonality_periods: np.ndarray, num_harmonics: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
  """Return unique Fourier frequencies for given periods and harmonics."""
  seasonality_periods = np.array(seasonality_periods, dtype=np.float32)
  if np.any((num_harmonics > seasonality_periods / 2)):
    raise ValueError('Harmonic cannot exceed half seasonal period.')
  if seasonality_periods.shape != num_harmonics.shape:
    raise ValueError('Number of seasonal periods and harmonics must be equal.')
  if len(num_harmonics.shape) != 1:
    raise ValueError(
        'Arguments `num_harmonics` and `seasonality_periods` must be rank 1.'
    )
  if seasonality_periods.shape[0] == 0:
    return (np.zeros(0), np.zeros(0))
  harmonics = [np.arange(1, h + 1, dtype=np.float32) for h in num_harmonics]
  frequencies = np.concatenate(
      [h / p for (h, p) in zip(harmonics, seasonality_periods)]
  )
  _, idx = np.unique(frequencies, return_index=True)
  idx_sort = np.sort(idx)
  unique_frequencies = frequencies[idx_sort]
  unique_harmonics = np.concatenate(harmonics)[idx_sort]
  return (unique_frequencies, unique_harmonics)


def make_seasonal_features(
    x: jax.typing.ArrayLike,
    seasonality_periods: np.ndarray,
    num_harmonics: np.ndarray,
    rescale: bool = False,
) -> jnp.ndarray:
  """Returns a set of cos and sin features for each seasonality period."""
  x = jnp.reshape(x, (-1, 1))
  frequencies, harmonics = make_seasonal_frequencies(
      seasonality_periods, num_harmonics
  )
  y = 2 * jnp.pi * frequencies * x
  features = jnp.column_stack((jnp.cos(y), jnp.sin(y)))
  denominator = jnp.tile(harmonics, 2)
  return features / denominator if rescale else features


def make_fourier_features(
    x: jax.typing.ArrayLike, max_degree: int, rescale: bool = False
) -> jnp.ndarray:
  """Returns a set of sine and cosine basis functions."""
  x = jnp.reshape(x, (-1, 1))
  degrees = jnp.arange(max_degree)
  y = 2 * jnp.pi * 2**degrees * x
  features = jnp.column_stack((jnp.cos(y), jnp.sin(y)))
  denominator = jnp.tile(degrees + 1, 2)
  return features / denominator if rescale else features


prior_base_d = tfd.Logistic


def prior_model_fn(mlp_template):
  yield prior_base_d(0.0, 1.0)  # log_noise_scale
  yield prior_base_d(-1.5, 1.0)  # shape
  yield prior_base_d(0.0, 1.0)  # inflated_loc_probs

  leaves = jax.tree_util.tree_leaves(mlp_template)
  flat_mlp_params = []
  for p in leaves:
    weights = yield prior_base_d(0.0, jnp.ones_like(p))
    flat_mlp_params.append(weights)


def make_likelihood_model(
    params: jax.typing.ArrayLike,
    x: jax.Array,
    mlp: nn.Module,
    mlp_template: flax.core.frozen_dict.FrozenDict,
    distribution: str,
) -> tfd.Distribution:
  """Creates likelihood for the model.

  For a Gaussian likelihood, the likelihood is centered at the predictions and
  the scale is given by the noise scale.

  For a NegativeBinomial, we parametrize in terms of mean and shape
  following Salinas et al. "DeepAR: Probabilistic Forecasting with
  Autoregressive Recurrent Networks" https://arxiv.org/pdf/1704.04110.pdf
  (Section 3.1), but use a shared shape parameter, which is the inverse of the
  total count. The variance equals `mean + mean^2 * shape`.

  See also https://en.wikipedia.org/wiki/Negative_binomial_distribution, in
  particular the section on Alternative parameterizations, which defines:

  ```
  probs = mean / variance
  total_count = mean^2 / (variance - mean)
  ```

  so that `variance = mean + mean^2 / total_count`
  and thus `total_count = 1 / shape`.

  From this we get the parameters passed to `tfd.NegativeBinomial` below:

  ```
  total_count = 1 / shape
  probs = mean / (mean + mean^2 * shape)
        = 1 / (1 + mean * shape)
  ```

  For zero inflation we follow ideas from https://arxiv.org/pdf/2010.09647.pdf

  Args:
    params: Pytree of particles.
    x: Time indices for the training set.
    mlp: BayesianNeuralField1D model instance.
    mlp_template: MLP variables.
    distribution: Either 'NORMAL', 'NB' (Negative Binomial) or 'ZINB' (Zero
      Inflated Negative Binomial) are supported.

  Returns:
    A distribution with the likelihood for each sample in the batch.
  """
  if LikelihoodDist(distribution) == LikelihoodDist.NORMAL:
    log_noise_scale = params[0]
    treedef = jax.tree_util.tree_structure(mlp_template)
    mlp_params = jax.tree_util.tree_unflatten(treedef, params[3:])
    predictions = mlp.apply(mlp_params, x)
    return tfd.Independent(
        tfd.Normal(predictions, 0.01 + jnp.exp(log_noise_scale)), 1
    )

  elif LikelihoodDist(distribution) == LikelihoodDist.NB:
    treedef = jax.tree_util.tree_structure(mlp_template)
    mlp_params = jax.tree_util.tree_unflatten(treedef, params[3:])
    predictions = mlp.apply(mlp_params, x)
    mean = jax.nn.softplus(predictions)
    shape = jax.nn.softplus(params[1])

    negative_binomial = tfd.NegativeBinomial(
        total_count=1 / shape, logits=-jnp.log(shape) - jnp.log(mean)
    )
    return tfd.Independent(negative_binomial, 1)

  elif LikelihoodDist(distribution) == LikelihoodDist.ZINB:
    treedef = jax.tree_util.tree_structure(mlp_template)
    mlp_params = jax.tree_util.tree_unflatten(treedef, params[3:])
    predictions = mlp.apply(mlp_params, x)
    mean = jax.nn.softplus(predictions)
    shape = jax.nn.softplus(params[1])
    inflated_loc_probs = 1 / (1 + jnp.exp(-params[2]))

    zero_inflated_negative_binomial = tfd.ZeroInflatedNegativeBinomial(
        total_count=1 / shape,
        logits=-jnp.log(shape) - jnp.log(mean),
        inflated_loc_probs=inflated_loc_probs * jnp.ones(shape=mean.shape),
    )
    return tfd.Independent(zero_inflated_negative_binomial, 1)

  else:
    raise AssertionError('Unknown likelihood distribution:', distribution)


class BayesianNeuralField1D(nn.Module):
  """Linen Module implementing a 1D Bayesian neural field."""

  width: int
  depth: int
  input_scales: np.ndarray
  fourier_degrees: np.ndarray
  interactions: np.ndarray
  num_seasonal_harmonics: np.ndarray = flax.struct.field(
      default_factory=lambda: np.zeros((0,))
  )
  seasonality_periods: np.ndarray = flax.struct.field(
      default_factory=lambda: np.zeros((0,))
  )

  @nn.compact
  def __call__(self, x):
    init = nn.initializers.normal(1.0)

    if len(x.shape) == 1:
      x = x[..., jnp.newaxis]
    log_scale_adjustment = self.param(
        'log_scale_adjustment', init, x.shape[-1:]
    )
    scaled_x = x / (self.input_scales * jnp.exp(log_scale_adjustment))

    seasonal_features = make_seasonal_features(
        x[..., 0],
        self.seasonality_periods,
        self.num_seasonal_harmonics,
        rescale=True,
    )

    fourier_features = [
        make_fourier_features(scaled_x[..., i], degree, rescale=True)
        for i, degree in enumerate(self.fourier_degrees)
        if degree > 0
    ]

    interaction_features = jnp.prod(scaled_x[:, self.interactions], axis=-1)

    def make_layer_scale(name, shape=()):
      inv_sp_layer_scale = self.param(name, init, shape)
      return jax.nn.softplus(inv_sp_layer_scale)

    features = [
        scaled_x,
        *fourier_features,
        seasonal_features,
        interaction_features,
    ]
    features = [
        f * jax.nn.softplus(self.param(f'feature_inv_sp_scale{i}', init, ()))
        for i, f in enumerate(features) if f.size > 0
    ]
    h = jnp.concatenate(features, -1)

    activation_weight = jax.nn.sigmoid(
        self.param('logit_activation_weight', init, ())
    )

    def activation_fn(x):
      return activation_weight * nn.elu(x) + (1 - activation_weight) * nn.tanh(
          x
      )

    for layer_id in range(self.depth):
      layer = nn.Dense(self.width, kernel_init=init, bias_init=init)
      layer_scale = make_layer_scale(f'inv_sp_layer_scale{layer_id}')
      # Equivalent to scaling the variance of the weight prior by 1/num_inputs.
      h = h / jnp.sqrt(jnp.shape(h)[-1])
      h = activation_fn(layer_scale * layer(h))
    output_layer = nn.Dense(1, kernel_init=init, bias_init=init)
    output_scale = make_layer_scale('inv_sp_output_scale')
    # Equivalent to scaling the variance of the weight prior by 1/num_inputs.
    h = h / jnp.sqrt(h.shape[-1])
    return output_scale * output_layer(h)[..., 0]



# Copyright 2024 The bayesnf Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model code for spatiotemporal Bayesian neural field."""
from collections.abc import Sequence

from bayesnf import inference
from bayesnf import models
import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
import tensorflow_probability.substrates.jax as tfp


tfd = tfp.distributions


def seasonality_to_float(seasonality: str, freq: str) -> float:
  """Convert a valid pandas frequency string to a float, relative `freq`.

  Internally, this computes the number of `freq`s in a long period (to
  account for leap years), and the number of `seasonality`s, and returns the
  ratio.

  For example
    seasonality_to_float('Y', 'D') == 365.25

    seasonality_to_float('Y', 'W') == 52.25

    seasonality_to_float('M', 'D') == 30.4375

  Args:
    seasonality: A valid pandas frequency.
    freq: A valid pandas frequency. Should be shorter than `seasonality`.

  Returns:
    A float of how many `seasonality` periods are in a `freq`, on average.
  """
  four_years = pd.date_range('2020-01-01', periods=5, freq='Y')
  y = four_years.to_period(seasonality)
  num_seasonality = (y[-1] - y[0]).n

  x = pd.date_range(y[0].start_time, y[-1].start_time).to_period(freq)
  num_freq = (x[-1] - x[0]).n

  return num_freq / num_seasonality


def seasonalities_to_array(
    seasonalities: Sequence[float | str], freq: str
) -> np.ndarray:
  """Convert a list of floats or strings to durations relative to a frequency.

  Args:
    seasonalities: A list of floats or strings representing durations relative
      to `freq`. For example, if `freq` is 'D', either 365 or 'Y' would
      represent a year. If `freq` is 'M', then either 12 or 'Y' represents a
      year.
    freq: Frequency of the data.

  Raises:
    TypeError: If the seasonality is less than or equal to 1.

  Returns:
    Array of floats greater than 1, representing seasonalities.
  """
  ret = []
  for seasonality in seasonalities:
    if isinstance(seasonality, str):
      seasonality_float = seasonality_to_float(seasonality, freq)
      if seasonality_float < 1:
        raise TypeError(f'{seasonality=} should represent a time '
                        f'span greater than {freq=}, but {seasonality} '
                        f'is {seasonality_float:.2f} of a {freq}')

    else:
      seasonality_float = seasonality
      if seasonality_float < 1:
        raise TypeError(f'{seasonality_float=} should be larger than 1.')
    ret.append(seasonality_float)
  return np.array(ret)


def _convert_datetime_col(table, time_column, timetype, freq, time_min=None):
  if timetype == 'index':
    first_date = pd.to_datetime('2020-01-01').to_period(freq)
    table[time_column] = table[time_column].dt.to_period(freq)
    table[time_column] = (table[time_column] - first_date).apply(lambda x: x.n)
  if time_min is None:
    time_min = table[time_column].min()
  table[time_column] = table[time_column] - time_min
  return table, time_min


class SpatiotemporalDataHandler:
  """Base class for preparing spatiotemporal data."""

  def __init__(
      self,
      feature_cols: Sequence[str],
      target_col: str,
      timetype: str,
      freq: str,
      standardize: Sequence[str] | None = None,
  ):
    self.feature_cols = feature_cols
    self.target_col = target_col
    self.timetype = timetype
    self.freq = freq
    self.standardize = standardize
    self.mu_ = None
    self.std_ = None
    self.time_min_ = None
    self.time_scale_ = None

  @property
  def _time_idx(self) -> int:
    return 0

  @property
  def _time_column(self) -> str:
    return self.feature_cols[self._time_idx]

  def get_target(self, table: pd.DataFrame) -> np.ndarray:
    table = self._maybe_filter_target_nans(table)
    return table[self.target_col].values

  def _maybe_filter_target_nans(self, table: pd.DataFrame) -> pd.DataFrame:
    if self.target_col in table.columns:
      return table[table[self.target_col].notna()]
    return table

  def copy_and_filter_table(self, table: pd.DataFrame) -> pd.DataFrame:
    return self._maybe_filter_target_nans(table.copy())

  def get_train(self, table: pd.DataFrame) -> np.ndarray:
    """Fetch training data."""
    table = self.copy_and_filter_table(table)
    self.mu_ = np.zeros(len(self.feature_cols))
    self.std_ = np.ones(len(self.feature_cols))

    table, self.time_min_ = _convert_datetime_col(
        table, self._time_column, self.timetype, self.freq, None)
    features = table[self.feature_cols].values
    self.time_scale_ = features[:, self._time_idx].max()

    if self.standardize:
      if self._time_column in self.standardize:
        raise TypeError('Do not standardize the time column!')
      idx = [self.feature_cols.index(f) for f in self.standardize]
      self.mu_[idx] = np.mean(features[:, idx].astype(float), axis=0)
      self.std_[idx] = np.std(features[:, idx].astype(float), axis=0)
      features = (features - self.mu_) / self.std_

    return features

  def get_test(self, table: pd.DataFrame) -> np.ndarray:
    """Fetch testing data. Call this after `get_train`."""
    table = self.copy_and_filter_table(table)
    table, _ = _convert_datetime_col(
        table, self._time_column, self.timetype, self.freq, self.time_min_)

    features = table[self.feature_cols].values

    if self.standardize:
      features = (features - self.mu_) / self.std_

    return features

  def get_input_scales(self) -> np.ndarray:
    input_scales = np.ones(len(self.feature_cols))
    input_scales[self._time_idx] = self.time_scale_
    return input_scales


class BayesianNeuralFieldEstimator:
  """Base class for BNF estimator."""

  _ensemble_dims: int
  _prior_weight: float = 1.0
  _scale_epochs_by_batch_size: bool = False

  def __init__(
      self,
      num_seasonal_harmonics: Sequence[int],
      seasonality_periods: Sequence[float | str],
      feature_cols: Sequence[str],
      target_col: str,
      timetype: str,
      freq: str,
      fourier_degrees: Sequence[float] | None = None,
      observation_model: str = 'NORMAL',
      depth: int = 2,
      width: int = 512,
      standardize: Sequence[str] | None = None,
      interactions: Sequence[tuple[int, int]] | None = None,
  ):
    self.num_seasonal_harmonics = num_seasonal_harmonics
    self.seasonality_periods = seasonality_periods
    self.observation_model = observation_model
    self.depth = depth
    self.width = width
    self.feature_cols = feature_cols
    self.target_col = target_col
    self.timetype = timetype
    self.freq = freq
    self.fourier_degrees = fourier_degrees
    self.standardize = standardize
    self.interactions = interactions

    self.losses_ = None
    self.params_ = None
    self.data_handler = SpatiotemporalDataHandler(
        self.feature_cols,
        self.target_col,
        self.timetype,
        self.freq,
        standardize=self.standardize)

  def _get_fourier_degrees(self, batch_shape: tuple[int, ...]) -> np.ndarray:
    """Set default fourier degrees, or verify shape is correct."""
    if self.fourier_degrees is None:
      fourier_degrees = np.full(batch_shape[-1], 5, dtype=int)
    else:
      fourier_degrees = np.atleast_1d(self.fourier_degrees).astype(int)
      if fourier_degrees.shape[-1] != batch_shape[-1]:
        raise ValueError(
            'The length of fourier_degrees ({}) must match the '
            'input dimension dimension ({}).'.format(
                fourier_degrees.shape[-1], batch_shape[-1]
            )
        )
    return fourier_degrees

  def _get_interactions(self) -> np.ndarray:
    """Set default fourier degrees, or verify shape is correct."""
    if self.interactions is None:
      interactions = np.zeros((0, 2), dtype=int)
    else:
      interactions = np.array(self.interactions).astype(int)
      if np.ndim(interactions) != 2 or interactions.shape[-1] != 2:
        raise ValueError(
            'The argument for `interactions` should be a 2-d array of integers '
            'of shape (N, 2), indicating the column indices to interact (the '
            f' passed shape was {interactions.shape})')
    return interactions

  def _model_args(self, batch_shape):
    return {
        'depth': self.depth,
        'input_scales': self.data_handler.get_input_scales(),
        'num_seasonal_harmonics': np.array(self.num_seasonal_harmonics),
        'seasonality_periods': seasonalities_to_array(
            self.seasonality_periods, self.freq),
        'width': self.width,
        'init_x': batch_shape,
        'fourier_degrees': self._get_fourier_degrees(batch_shape),
        'interactions': self._get_interactions(),
    }

  def predict(self, table, quantiles=(0.5,)):
    test_data = self.data_handler.get_test(table)
    return inference.predict_bnf(
        test_data,
        self.observation_model,
        params=self.params_,
        model_args=self._model_args(test_data.shape),
        quantiles=quantiles,
        ensemble_dims=self._ensemble_dims,
        approximate_quantiles=False,
    )

  def fit(self, table, seed):
    raise NotImplementedError('Should be implemented by subclass')

  def likelihood_model(self, table: pd.DataFrame) -> tfd.Distribution:
    """Access the likelihood distribution after calling `.fit`."""
    test_data = self.data_handler.get_test(table)
    mlp, mlp_template = inference.make_model(
        **self._model_args(test_data.shape))
    for _ in range(self._ensemble_dims - 1):
      mlp.apply = jax.vmap(mlp.apply, in_axes=(0, None))
    mlp.apply = jax.pmap(mlp.apply, in_axes=(0, None))

    # This allows the likelihood to broadcast correctly with the batch of
    # predictions.
    params = self.params_._replace(**{  # pytype: disable=attribute-error
        self.params_._fields[i]: self.params_[i][..., jnp.newaxis]  # pytype: disable=unsupported-operands,attribute-error
        for i in range(3)})

    return models.make_likelihood_model(
        params,
        jnp.array(test_data),
        mlp,
        mlp_template,
        self.observation_model)


class BayesianNeuralFieldMAP(BayesianNeuralFieldEstimator):
  """Fit BNF using MAP estimation."""

  _ensemble_dims = 2

  def fit(
      self,
      table,
      seed,
      ensemble_size=16,
      learning_rate=0.005,
      num_epochs=5_000,
      batch_size=None,
      num_splits=1,
  ):
    train_data = self.data_handler.get_train(table)
    train_target = self.data_handler.get_target(table)
    if batch_size is None:
      batch_size = train_data.shape[0]
    if self._scale_epochs_by_batch_size:
      num_epochs = num_epochs * (train_data.shape[0] // batch_size)
    model_args = self._model_args((batch_size, train_data.shape[-1]))
    self.params_, self.losses_ = inference.fit_map(
        train_data,
        train_target,
        seed=seed,
        observation_model=self.observation_model,
        model_args=model_args,
        num_particles=ensemble_size,
        learning_rate=learning_rate,
        num_epochs=num_epochs,
        prior_weight=self._prior_weight,
        batch_size=batch_size,
        num_splits=num_splits)
    return self


class BayesianNeuralFieldMLE(BayesianNeuralFieldMAP):
  """Fit BNF using MLE estimation."""

  _prior_weight = 0.0

  def predict(self, table, quantiles=(0.5,)):
    test_data = self.data_handler.get_test(table)
    return inference.predict_bnf(
        test_data,
        self.observation_model,
        params=self.params_,
        model_args=self._model_args(test_data.shape),
        quantiles=quantiles,
        ensemble_dims=self._ensemble_dims,
        approximate_quantiles=True,
    )


class BayesianNeuralFieldVI(BayesianNeuralFieldEstimator):
  """Fit BNF using VI estimation."""

  _ensemble_dims = 3
  _scale_epochs_by_batch_size = True

  def fit(
      self,
      table,
      seed,
      ensemble_size=16,
      learning_rate=0.01,
      num_epochs=1_000,
      sample_size=5,
      kl_weight=0.1,
      batch_size=None,
  ):
    train_data = self.data_handler.get_train(table)
    train_target = self.data_handler.get_target(table)
    if batch_size is None:
      batch_size = train_data.shape[0]
    if self._scale_epochs_by_batch_size:
      num_epochs = num_epochs * (train_data.shape[0] // batch_size)
    model_args = self._model_args((batch_size, train_data.shape[-1]))
    _, self.losses_, self.params_ = inference.fit_vi(
        train_data,
        train_target,
        seed=seed,
        observation_model=self.observation_model,
        model_args=model_args,
        ensemble_size=ensemble_size,
        learning_rate=learning_rate,
        num_epochs=num_epochs,
        sample_size=sample_size,
        kl_weight=kl_weight,
        batch_size=batch_size,
    )
    return self



# Copyright 2024 The bayesnf Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset configuration in spatiotemporal experiments."""

import numpy

DATASET_CONFIG = {
    'air_quality': {
        'num_series': 12,
        'target_col': 'pm10',
        'timetype': 'index',
        'freq': 'H',
        'feature_cols': ['datetime', 'latitude', 'longitude'],
        'standardize': ['latitude', 'longitude'],
        'series_id_fmt': str,
    },
    'wind': {
        'num_series': 12,
        'target_col': 'wind',
        'timetype': 'index',
        'freq': 'D',
        'feature_cols': ['datetime', 'latitude', 'longitude'],
        'standardize': ['latitude', 'longitude'],
        'series_id_fmt': str,
    },
    'air': {
        'num_series': 12,
        'target_col': 'pm10',
        'timetype': 'index',
        'freq': 'D',
        'feature_cols': ['datetime', 'latitude', 'longitude'],
        'standardize': ['latitude', 'longitude'],
        'series_id_fmt': str,
    },
    'chickenpox': {
        'num_series': 12,
        'target_col': 'chickenpox',
        'timetype': 'index',
        'freq': 'W',
        'feature_cols': ['datetime', 'latitude', 'longitude'],
        'standardize': ['latitude', 'longitude'],
        'series_id_fmt': str,
    },
    'coprecip': {
        'num_series': 12,
        'target_col': 'ppt',
        'timetype': 'index',
        'freq': 'M',
        'feature_cols': ['datetime', 'latitude', 'longitude'],
        'standardize': ['latitude', 'longitude'],
        'series_id_fmt': str,
    },
    'sst': {
        'num_series': 12,
        'target_col': 'sst',
        'timetype': 'index',
        'freq': 'M',
        'feature_cols': ['datetime', 'latitude', 'longitude', 'soi'],
        'standardize': ['latitude', 'longitude'],
        'series_id_fmt': str,
    },
    'M3Month': {
        'num_series': 1428,
        'target_col': 'value',
        'timetype': 'index',
        'freq': 'M',
        'feature_cols': ['datetime'],
        'standardize': [],
        'series_id_fmt': lambda s: str(1402 + s),
    },
}


def _get_model_config():
  """Return model configs."""
  ret = {}

  ret['air_quality'] = {
      'map': {
          'width': 512,
          'depth': 2,
          'seasonality_periods': numpy.asarray([24, 24 * 7]),
          'num_seasonal_harmonics': numpy.asarray([4, 4]),
          'observation_model': 'NORMAL',
      }
  }
  ret['air_quality']['mle'] = ret['air_quality']['map']
  ret['air_quality']['vi'] = ret['air_quality']['map'] | {
      'width': 512,
      'observation_model': 'NORMAL',
  }

  ret['wind'] = {
      'map': {
          'width': 512,
          'depth': 2,
          'seasonality_periods': numpy.asarray([7, 365.25 / 12, 365.25]),
          'num_seasonal_harmonics': numpy.asarray([3, 10, 10]),
          'observation_model': 'NORMAL',
      }
  }
  ret['wind']['mle'] = ret['wind']['map']
  ret['wind']['vi'] = ret['wind']['map'] | {
      'observation_model': 'NORMAL',
  }

  ret['air'] = {
      'map': {
          'width': 512,
          'depth': 2,
          'seasonality_periods': numpy.asarray([7, 365.25 / 12, 365.25]),
          'num_seasonal_harmonics': numpy.asarray([3, 10, 10]),
          'observation_model': 'NORMAL',
      }
  }
  ret['air']['mle'] = ret['air']['map']
  ret['air']['vi'] = ret['air']['map'] | {
      'depth': 2,
      'observation_model': 'NORMAL',
  }

  ret['chickenpox'] = {
      'map': {
          'width': 256,
          'depth': 2,
          'seasonality_periods': numpy.asarray([
              4.0,
              52.1775,
          ]),
          'num_seasonal_harmonics': numpy.asarray([2.0, 10]),
          'observation_model': 'NORMAL',
      }
  }
  ret['chickenpox']['mle'] = ret['chickenpox']['map']
  ret['chickenpox']['vi'] = ret['chickenpox']['map'] | {
      'observation_model': 'NORMAL',
  }

  ret['coprecip'] = {
      'map': {
          'width': 512,
          'depth': 2,
          'seasonality_periods': numpy.asarray([12]),
          'num_seasonal_harmonics': numpy.asarray([6]),
          'observation_model': 'NORMAL',
      }
  }
  ret['coprecip']['mle'] = ret['coprecip']['map']
  ret['coprecip']['vi'] = ret['coprecip']['map']

  ret['sst'] = {
      'map': {
          'width': 768,
          'depth': 2,
          'seasonality_periods': numpy.asarray([
              12,
          ]),
          'num_seasonal_harmonics': numpy.asarray([
              6,
          ]),
          'observation_model': 'NORMAL',
      }
  }
  ret['sst']['mle'] = ret['sst']['map']
  ret['sst']['vi'] = ret['sst']['map']

  ret['M3Month'] = {
      'map': {
          'width': 1024,
          'depth': 2,
          'seasonality_periods': numpy.asarray([12]),
          'num_seasonal_harmonics': numpy.asarray([6]),
      }
  }
  ret['M3Month']['mle'] = ret['M3Month']['map']

  return ret


MODEL_CONFIG = _get_model_config()



# Copyright 2024 The bayesnf Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluate ST-BNF."""

import json
import os
import time
from typing import Any, Union

from bayesnf import spatiotemporal
import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
import tensorflow_probability.substrates.jax as tfp


tfd = tfp.distributions
ArrayT = jax.Array | np.ndarray


def drop_nan(x: ArrayT, y: ArrayT) -> tuple[ArrayT, ArrayT]:
  """Drop elements of x and y at indexes where y is NaN."""
  keep = ~np.isnan(y)
  return x[keep], y[keep]


def run_experiment(
    dataset: str,
    data_root: str,
    series_id: Union[int, str],
    output_dir: str,
    objective: str,
    dataset_config: dict[str, Any],
    model_config: dict[str, Any],
    inference_config: dict[str, Any],
    seed: jax.Array,
) -> tuple[ArrayT, ArrayT, ArrayT]:
  """Runs a single training experiment, writes output to output_dir."""
  path_train = os.path.join(data_root, f'{dataset}.{series_id}.train.csv')
  df_train = pd.read_csv(path_train, index_col=0, parse_dates=['datetime'])
  path_test = os.path.join(data_root, f'{dataset}.{series_id}.test.csv')
  df_test = pd.read_csv(path_test, index_col=0, parse_dates=['datetime'])

  os.makedirs(output_dir, exist_ok=True)
  path_model = os.path.join(
      output_dir, f'bnf-{objective}.{dataset}.{series_id}.json'
  )
  model_config.update(dict(
      feature_cols=dataset_config['feature_cols'],
      target_col=dataset_config['target_col'],
      timetype=dataset_config['timetype'],
      freq=dataset_config.get('freq', None),
      standardize=dataset_config.get('standardize', None),
  ))

  if objective == 'vi':
    base_cls = spatiotemporal.BayesianNeuralFieldVI
    objective_specific_inference_args = {
        'kl_weight': inference_config.get('kl_weight', 1.0),
        'sample_size': inference_config.get('sample_size', 10),
    }
  elif objective == 'map':
    base_cls = spatiotemporal.BayesianNeuralFieldMAP
    objective_specific_inference_args = {
        'num_splits': inference_config.get('num_particle_splits', 1),
    }
  elif objective == 'mle':
    base_cls = spatiotemporal.BayesianNeuralFieldMLE
    objective_specific_inference_args = {
        'num_splits': inference_config.get('num_particle_splits', 1),
    }
  else:
    raise ValueError(f'objective={objective}')

  start_time = time.perf_counter()
  inference_args = (
      dict(
          learning_rate=inference_config['learning_rate'],
          num_epochs=inference_config['num_epochs'],
          batch_size=inference_config.get('batch_size', None),
          ensemble_size=inference_config['num_particles'],
      )
      | objective_specific_inference_args
  )

  model = base_cls(**model_config).fit(df_train, seed, **inference_args)

  df_train_and_test = pd.concat([df_train, df_test])
  means, quantiles = model.predict(df_train_and_test,
                                   quantiles=(0.5, 0.025, 0.975))
  losses = model.losses_
  assert losses is not None
  runtime = time.perf_counter() - start_time

  path_log = path_model.replace('.json', '.log.json')
  with open(path_log, 'w') as f:
    log = {
        'dataset': dataset,
        'series_id': series_id,
        'runtime': runtime,
        'objective': objective,
        'dataset_config': dataset_config,
        'model_config': model_config,
        'inference_config': inference_config,
    }
    json.dump(log, f, indent=2, default=repr)

  path_loss = path_model.replace('.json', '.loss.csv')
  df_log = pd.DataFrame(losses.reshape((-1, losses.shape[-1])).T)
  with open(path_loss, 'w') as f:
    df_log.to_csv(f, index=False)

  pred_index = model.data_handler.copy_and_filter_table(df_train_and_test).index
  df_pred = pd.DataFrame(
      {
          'yhat': np.mean(means, axis=range(len(means.shape) - 1)),
          'yhat_p50': quantiles[0],
          'yhat_lower': quantiles[1],
          'yhat_upper': quantiles[2],
      },
      index=pred_index,
  )
  df_pred.sort_index(inplace=True)
  path_pred = path_model.replace('.json', '.pred.csv')
  with open(path_pred, 'w') as f:
    df_pred.to_csv(f, index=True)

  return losses, means, jnp.array(quantiles)



# Copyright 2024 The bayesnf Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Run BNF time-series analysis experiments on M3 dataset."""

from typing import Sequence

from absl import app
from absl import flags
from absl import logging
from bayesnf import dataset_config
from bayesnf import evaluate
import jax


_OUTPUT_DIR = flags.DEFINE_string(
    'output_dir', None, 'Output directory.', required=True
)

_DATA_ROOT = flags.DEFINE_string(
    'data_root', None, 'Location of input data.', required=True
)

_DATASET = flags.DEFINE_enum(
    'dataset',
    None,
    enum_values=dataset_config.DATASET_CONFIG.keys(),
    help='Dataset name',
    required=True,
)

_OBJECTIVE = flags.DEFINE_enum(
    'objective',
    default='map',
    enum_values=['map', 'mle', 'vi'],
    help='Training objective',
)

_START_ID = flags.DEFINE_integer(
    'start_id', 0, 'Run experiments on series with IDs >= this value.'
)

_STOP_ID = flags.DEFINE_integer(
    'stop_id', None, 'Run experiments on series with IDs < this value.'
)

_NUM_PARTICLES = flags.DEFINE_integer(
    'num_particles', None, 'Override the number of particles for inference.'
)


# pylint: disable=bad-whitespace
def _get_inference_config():
  """Return inference configs."""
  ret = {}

  ret['air_quality'] = {
      'map': {
          'num_particles': 16,
          'num_epochs': 4000,
          'learning_rate': 0.005,
          'batch_size': 38096,
      },
      'vi': {
          'num_particles': 16,
          'num_epochs': 500,
          'learning_rate': 0.01,
          'batch_size': 3500,
          'kl_weight': 0.2,
          'sample_size': 5,
      },
  }
  ret['air_quality']['mle'] = ret['air_quality']['map']

  ret['wind'] = {
      'map': {
          'num_particles': 64,
          'num_epochs': 10000,
          'learning_rate': 0.005,
      },
      'vi': {
          'num_particles': 64,
          'num_epochs': 2000,
          'learning_rate': 0.01,
          'batch_size': 3944,
          'kl_weight': 0.1,
          'sample_size': 5,
      },
  }
  ret['wind']['mle'] = ret['wind']['map']

  ret['air'] = {
      'map': {
          'num_particles': 8,
          'num_epochs': 7500,
          'learning_rate': 0.005,
      },
      'vi': {
          'num_particles': 8,
          'num_epochs': 1000,
          'learning_rate': 0.01,
          'batch_size': 3800,
          'kl_weight': 0.2,
          'sample_size': 5,
      },
  }
  ret['air']['mle'] = ret['air']['map']

  ret['chickenpox'] = {
      'map': {
          'num_particles': 64,
          'num_epochs': 10000,
          'learning_rate': 0.005,
      },
      'vi': {
          'num_particles': 64,
          'num_epochs': 1000,
          'learning_rate': 0.01,
          'batch_size': 511,
          'kl_weight': 0.1,
          'sample_size': 5,
      },
  }
  ret['chickenpox']['mle'] = ret['chickenpox']['map']

  ret['coprecip'] = {
      'map': {
          'num_particles': 16,
          'num_epochs': 7500,
          'learning_rate': 0.005,
      },
      'vi': {
          'num_particles': 16,
          'num_epochs': 750,
          'learning_rate': 0.01,
          'batch_size': 3300,
          'kl_weight': 0.2,
          'sample_size': 5,
      },
  }
  ret['coprecip']['mle'] = ret['coprecip']['map']

  ret['sst'] = {
      'map': {
          'num_particles': 16,
          'num_epochs': 5000,
          'learning_rate': 0.005,
          'batch_size': 221127,
      },
      'vi': {
          'num_particles': 16,
          'num_epochs': 600,
          'learning_rate': 0.005,
          'batch_size': 8845,
          'kl_weight': 0.5,
          'sample_size': 5,
      },
  }
  ret['sst']['mle'] = ret['sst']['map']

  ret['M3Month'] = {
      'map': {
          'num_particles': 64,
          'num_epochs': 5000,
          'learning_rate': 0.01,
      },
  }
  ret['M3Month']['mle'] = ret['M3Month']['map']

  return ret


# pylint: enable=bad-whitespace

INFERENCE_CONFIG = _get_inference_config()


def main(argv: Sequence[str]):
  if len(argv) > 3:
    raise app.UsageError('Too many command-line arguments.')

  if _NUM_PARTICLES.value:
    for k in INFERENCE_CONFIG:
      for obj in INFERENCE_CONFIG[k]:
        INFERENCE_CONFIG[k][obj]['num_particles'] = _NUM_PARTICLES.value

  stop_id = (
      _STOP_ID.value
      or dataset_config.DATASET_CONFIG[_DATASET.value]['num_series']
  )
  for series_id in range(_START_ID.value, stop_id):
    logging.info('%s series_id %d', _DATASET.value, series_id)
    evaluate.run_experiment(
        dataset=_DATASET.value,
        data_root=_DATA_ROOT.value,
        series_id=dataset_config.DATASET_CONFIG[_DATASET.value][
            'series_id_fmt'](series_id),
        output_dir=_OUTPUT_DIR.value,
        objective=_OBJECTIVE.value,
        dataset_config=dataset_config.DATASET_CONFIG[_DATASET.value],
        model_config=dataset_config.MODEL_CONFIG[_DATASET.value][
            _OBJECTIVE.value
        ],
        inference_config=INFERENCE_CONFIG[_DATASET.value][_OBJECTIVE.value],
        seed=jax.random.PRNGKey(2023100400 + int(series_id)),
    )


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 The bayesnf Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Launch spatiotemporal BNF experiments."""

from absl import app
from absl import flags

from xmanager import xm
from xmanager import xm_local


DATASET_SIZE = {
    'air_quality': 12,
    'wind': 12,
    'air': 12,
    'chickenpox': 12,
    'coprecip': 12,
    'sst': 12,
    'M3Month': 1428,
}

_OUTPUT_DIR = flags.DEFINE_string(
    'output_dir', None, 'Output directory.', required=True)

_START_ID = flags.DEFINE_integer(
    'start_id', 0, 'Run experiments on series with IDs >= this value.')

_STOP_ID = flags.DEFINE_integer(
    'stop_id', -1, 'Run experiments on series with IDs < this value.')

_DATASET = flags.DEFINE_enum(
    'dataset',
    default=None,
    enum_values=list(DATASET_SIZE),
    help='Dataset name',
    required=True,
)

_OBJECTIVE = flags.DEFINE_enum(
    'objective',
    default='map',
    enum_values=['map', 'mle', 'vi'],
    help='Training objective',
)

_NUM_WORK_UNITS = flags.DEFINE_integer(
    'num_work_units', 1, 'Divide the experiments into this many work units.'
)

_NUM_PARTICLES = flags.DEFINE_integer(
    'num_particles', None, 'Override the number of particles for inference.')


def main(_):
  with xm_local.create_experiment(
      experiment_title=f'Spatiotemporal BNF-{_OBJECTIVE.value} {_DATASET.value}'
  ) as experiment:
    # Packaging.
    [executable] = experiment.package([
        xm.bazel_binary(
            label='//research/probability/bnf/opensource:runner',
            executor_spec=xm_local.Vertex.Spec(),
        ),
    ])

    # Execution.
    requirements = xm.JobRequirements(tpu_v5='2x2')

    # Start and stop ID.
    start_id = _START_ID.value
    stop_id = _STOP_ID.value
    if (stop_id == - 1) or (stop_id > DATASET_SIZE[_DATASET.value]):
      stop_id = DATASET_SIZE[_DATASET.value]
    assert start_id < stop_id, f'No experiments: ({start_id=}, {stop_id=}).'

    # Split among work units.
    splits = [
        (i * (stop_id - start_id)) // _NUM_WORK_UNITS.value + start_id
        for i in range(_NUM_WORK_UNITS.value + 1)
    ]
    for i in range(_NUM_WORK_UNITS.value):
      experiment.add(
          xm.Job(
              executable,
              args={
                  'output_dir': _OUTPUT_DIR.value,
                  'dataset': _DATASET.value,
                  'start_id': splits[i],
                  'stop_id': splits[i+1],
                  'objective': _OBJECTIVE.value,
                  'num_particles': _NUM_PARTICLES.value,
              },
              executor=xm_local.Vertex(requirements=requirements)))


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 The bayesnf Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code to run inference using spatiotemporal BNF."""

import functools
from typing import Any, Callable

from bayesnf import models
import flax
import jax
from jax import numpy as jnp
from jaxtyping import PyTree  # pylint: disable=g-importing-member
import numpy as np
import optax
from tensorflow_probability.substrates import jax as tfp  # pylint: disable=g-importing-member


tfd = tfp.distributions
ArrayT = jax.Array | np.ndarray


def permute_dataset(
    features: ArrayT, target: ArrayT, seed: jax.Array
) -> tuple[ArrayT, ArrayT]:
  permutation = jax.random.permutation(seed, jnp.arange(target.shape[0]))
  return features[permutation], target[permutation]


@functools.partial(jax.jit, static_argnames=('axis',))
def _normal_quantile_via_root(means, scales, q, axis=(0, 1)):
  n = tfd.Normal(means, scales)
  res = tfp.math.find_root_chandrupatla(
      lambda x: n.cdf(x).mean(axis) - q,
      low=jnp.amin(means) - 5 * jnp.amax(scales),
      high=jnp.amax(means) + 5 * jnp.amax(scales),
      value_tolerance=1e-5,
      max_iterations=60,
  )
  return res.estimated_root


@functools.partial(jax.jit, static_argnames=('axis',))
def _approximate_normal_quantile(
    means: jax.Array, scales: jax.Array, q: float, axis=(0, 1)
) -> jax.Array:
  """Fast approximate quantile for a mixture of gaussians.

  Compute the mean and standard deviation of the mixture distribution,
  then pretend the mixture distribution is a Normal distribution and
  compute Normal quantiles, as in [1].

  [1] Lakshminarayanan, Pritzel, and Blundell. "Simple and scalable
      predictive uncertainty estimation using deep ensembles." NeurIPS 2017.

  Args:
    means: Array of normal means
    scales: Array of normal scales. Should broadcast with `means`.
    q: Quantile to compute in (0, 1).
    axis: Batch axis to reduce over.

  Returns:
    Array of approximate `q`th quantiles, which should have
    shape `means.shape[-1]`.
  """
  mixture_mean = means.mean(axis)
  mixture_scale = jnp.sqrt(
      (jnp.square(scales) + jnp.square(means)).mean(axis)
      - jnp.square(mixture_mean)
  )
  n = tfd.Normal(mixture_mean, mixture_scale)
  return n.quantile(q)


def _get_percentile_normal(
    means, scales, quantiles, axis=(0, 1), approximate=False
) -> list[jax.Array]:
  """Compute the normal percentiles for a mixture."""
  if approximate:
    quantile_fn = _approximate_normal_quantile
  else:
    quantile_fn = _normal_quantile_via_root
  forecast_quantiles = []
  for q in quantiles:
    forecast_quantiles.append(
        quantile_fn(means, scales[..., jnp.newaxis], q, axis)
    )
  return forecast_quantiles


def _make_forecast_inner(model_args, distribution):
  """Construct inner forecast function for MAP and VI."""

  def forecast_inner(params, x_subset):
    likelihood = models.make_likelihood_model(
        params, x_subset, *make_model(**model_args), distribution
    )
    if distribution == models.LikelihoodDist.NORMAL:
      return (likelihood.distribution.loc, likelihood.distribution.scale)
    elif distribution == models.LikelihoodDist.NB:
      return (
          likelihood.distribution.total_count,
          likelihood.distribution.logits,
      )
    elif distribution == models.LikelihoodDist.ZINB:
      return (
          likelihood.distribution.total_count,
          likelihood.distribution.logits,
          likelihood.distribution.inflated_loc_probs,
      )
    else:
      raise TypeError('Distribution must be one of NORMAL, NB, or ZINB.')

  return forecast_inner


def forecast_parameters_batched(
    features: jax.Array,
    params: PyTree,
    distribution: models.LikelihoodDist,
    forecast_inner: Callable[[PyTree, jax.Array], PyTree],
    batchsize: int = 1024,
) -> tuple[jax.Array, ...]:
  """Computes parameters of the likelihood distribution.

  To handle large datasets, this operation is batched.

  Args:
    features: Input data to compute parameters over.
    params: Pytree of model parameters.
    distribution: Name of likelihood distribution.
    forecast_inner: Inner function used for forecasting. Use
      `_make_forecast_inner` to construct.
    batchsize: Batch size to use for splitting up the computation.

  Returns:
    A tuple of parameters, depending on the likelihood distribution.
    Normal likelihood: `loc` of shape BATCH_SHAPE+OBS_SHAPE and `scale` of
      shape BATCH_SHAPE
    Negative Binomial: `total_count` of shape BATCH_SHAPE and `logits` of
      shape BATCH_SHAPE + OBS_SHAPE
    Zero Inflated Negative Binomial: `total_count` of shape BATCH_SHAPE,
      `logits` of shape BATCH_SHAPE + OBS_SHAPE, and  `inflated_loc_probs` of
      shape BATCH_SHAPE.
    Note that `loc` and `logits` have shape BATCH_SHAPE + OBS_SHAPE. Other
    parameters are shared across time and only have BATCH_SHAPE.
  """
  forecast_params_slices = [[], [], []]

  data_size = features.shape[0]
  num_batches = data_size // batchsize
  for i in range(num_batches + 1):
    if i == num_batches:
      # last batch can be a partial batch
      batch_slice = slice(i * batchsize, None)
      if not batch_slice:
        # empty batch
        continue
    else:
      batch_slice = slice(i * batchsize, (i + 1) * batchsize)
    forecast_params = forecast_inner(params, features[batch_slice])
    for i, fc_param in enumerate(forecast_params):
      # This loop can be over 2 params for NORMAL and NB or 3 params for ZINB:
      # (count, logit) or (loc, scale), or (count, logit, inflated_loc_scale)
      forecast_params_slices[i].append(fc_param)

  # logit and loc are per sample so we concatenate across batches:
  if distribution == models.LikelihoodDist('NORMAL'):
    loc = jnp.concatenate(forecast_params_slices.pop(0), axis=-1)
    forecast_params = [k[0] for k in forecast_params_slices if k]
    forecast_params.insert(0, loc)
  elif distribution == models.LikelihoodDist('NB'):
    logit = jnp.concatenate(forecast_params_slices.pop(1), axis=-1)
    forecast_params = [k[0] for k in forecast_params_slices if k]
    forecast_params.insert(1, logit)
  elif distribution == models.LikelihoodDist('ZINB'):
    count = forecast_params_slices[0][0]
    logit = jnp.concatenate(forecast_params_slices[1], axis=-1)
    zero_mass = jnp.concatenate(forecast_params_slices[2], axis=-1)
    forecast_params = count, logit, zero_mass
  else:
    raise TypeError('Distribution must be NORMAL, NB, or ZINB.')

  return tuple(forecast_params)


def make_vi_init(prior_d: tfd.Distribution):
  """Given a JDS prior, construct a surrogate posterior init function."""
  xs = prior_d.sample(seed=jax.random.PRNGKey(0))

  def _fn():
    for i, x in enumerate(xs):
      # Surrogate mean.
      if len(x.shape) != 2:
        yield tfd.Deterministic(
            jnp.zeros_like(x),
            name=f'zero_initial_mean_for_bias_or_transformed_scale_{i}',
        )
      else:
        yield tfd.TruncatedNormal(
            0.0,
            jnp.ones_like(x),
            low=-2,
            high=2,
            name=f'initial_weight_matrix_{i}',
        )

      yield tfd.Deterministic(
          tfp.math.softplus_inverse(0.3) * jnp.ones_like(x),
          name=f'initial_inv_softplus_surrogate_scale_{i}',
      )

  return tfd.JointDistributionCoroutine(
      _fn, use_vectorized_map=True, batch_ndims=0
  ).sample


def make_model(
    width: int,
    depth: int,
    input_scales: np.ndarray,
    num_seasonal_harmonics: np.ndarray,
    seasonality_periods: np.ndarray,
    init_x: tuple[int, ...],
    fourier_degrees: np.ndarray,
    interactions: np.ndarray,
) -> tuple[models.BayesianNeuralField1D, flax.core.scope.FrozenVariableDict]:
  """Instantiate and initialize BayesianNeuralField1D model."""
  mlp = models.BayesianNeuralField1D(
      width=width,
      depth=depth,
      input_scales=input_scales,
      fourier_degrees=fourier_degrees,
      interactions=interactions,
      num_seasonal_harmonics=num_seasonal_harmonics,
      seasonality_periods=seasonality_periods,
  )
  mlp_template = mlp.init(
      jax.random.PRNGKey(0), jnp.zeros(init_x, dtype=jnp.float32)
  )

  return mlp, mlp_template


def make_prior(**kwargs: dict[str, Any]) -> tfd.JointDistributionCoroutine:
  kwargs.pop('likelihood_distribution', None)
  prior_d = tfd.JointDistributionCoroutine(
      functools.partial(models.prior_model_fn, make_model(**kwargs)[1]),
      use_vectorized_map=True,
      batch_ndims=0,
  )
  return prior_d


def _build_observation_distribution(distribution, forecast_params):
  """Returns (zero inflated) Negative Binomial distibution given parameters.

  Args:
    distribution: Indicates whether a zero-inflated models.LikelihoodDist.ZINB
      or models.LikelihoodDist.NB disribution should be returned.
    forecast_params: Tuple of total_count, logits, and optionally
      maybe_zero_mass
  """
  total_count, logits, *maybe_zero_mass = forecast_params
  if distribution == models.LikelihoodDist.NB:
    return tfd.NegativeBinomial(
        total_count=total_count[..., jnp.newaxis], logits=logits
    )

  elif distribution == models.LikelihoodDist.ZINB:
    inflated_loc_probs = maybe_zero_mass[0]

    return tfd.ZeroInflatedNegativeBinomial(
        total_count=total_count[..., jnp.newaxis],
        logits=logits,
        inflated_loc_probs=inflated_loc_probs,
    )
  else:
    raise ValueError(f'Unknown distribution: {distribution}')


@functools.partial(jax.jit, static_argnames=('ensemble_axes',))
def _get_nb_quantiles_root(
    dist: tfd.Distribution, q: float, ensemble_axes: tuple[int, ...] = (0, 1, 2)
) -> jax.Array:
  """Returns (zero inflated) Negative Binomial quantiles via root-finding.

  Quantiles are computed using root-finding. We post-process the result from
  `_nb_quantile_via_root` by taking the ceiling. Since the CDF is already
  above zero for an input of zero, we filter these out and manually set the
  values to zero. The upper limit for the root finding algorithm is based on
  the Chebyshev bound (http://en.wikipedia.org/wiki/Chebyshev%27s_inequality).

  Note that this is intended to be used with VI and MAP, and not AIS, where
  particles weights need to be incorporated into the average.

  Args:
    dist: Distribution, assumed to to be (zero inflated) Negative Binomial.
    q: Quantile to estimate.
    ensemble_axes: Int or tuple of ints that index ensemble dimensions. Default
      value is (0, 1, 2).
  """
  res = tfp.math.find_root_chandrupatla(
      lambda x: dist.cdf(x).mean(axis=ensemble_axes) - q,
      low=0.0,
      high=(
          jnp.amax(dist.mean())
          + 1.1 * jax.lax.rsqrt(1 - q) * jnp.amax(dist.stddev())
      ),
      value_tolerance=1e-5,
      max_iterations=60,
  )
  return jnp.ceil(
      jnp.where(
          dist.prob(0).mean(axis=ensemble_axes) > q, 0, res.estimated_root
      )
  )


def fit_vi(
    features: ArrayT,
    target: ArrayT,
    seed: jax.Array,
    observation_model: str,
    model_args: dict[str, Any],
    ensemble_size: int,
    learning_rate: float,
    num_epochs: int,
    sample_size: int,
    kl_weight: float,
    batch_size: int | None = None,
) -> tuple[
    tfp.distributions.JointDistributionSequential, jax.Array, jax.Array
]:
  """Fit BNF using an ensemble VI."""
  distribution = models.LikelihoodDist(observation_model)

  def _neg_energy_fn(params, x, y):
    return models.make_likelihood_model(
        params, x, *make_model(**model_args), distribution
    ).log_prob(y)

  return ensemble_vi(
      features,
      target,
      _neg_energy_fn,
      prior_d=make_prior(**model_args),
      ensemble_size=ensemble_size // jax.device_count(),
      learning_rate=learning_rate,
      num_epochs=num_epochs,
      seed=seed,
      sample_size=sample_size,
      kl_weight=kl_weight,
      batch_size=batch_size,
  )


def fit_map(
    features: ArrayT,
    target: ArrayT,
    seed: jax.Array,
    observation_model: str,
    model_args: dict[str, Any],
    num_particles: int,
    learning_rate: float,
    num_epochs: int,
    prior_weight: float = 1.0,
    batch_size: int | None = None,
    num_splits: int = 1,
) -> tuple[PyTree, np.ndarray]:
  """Fit a BNF using provided features and y data."""
  distribution = models.LikelihoodDist(observation_model)

  def _neg_energy_fn(params, x, y):
    return models.make_likelihood_model(
        params, x, *make_model(**model_args), distribution
    ).log_prob(y)

  target_scale = np.nanstd(target)

  def _make_init_fn(prior_d):
    xs = prior_d.sample(seed=jax.random.PRNGKey(0))

    def _fn():
      for i, x in enumerate(xs):
        if i == 0:
          # Initialize the noise scale (for the Normal observation distribution)
          # to be in the ballpark of the standard deviation of the observations.
          yield tfd.Deterministic(
              jnp.ones_like(x) * jnp.log(target_scale / 2.0),
              name='zero_initial_log_noise_scale',
          )
        elif len(x.shape) != 2:
          yield tfd.Deterministic(
              jnp.zeros_like(x),
              name=f'zero_initial_mean_for_bias_or_transformed_scale_{i}',
          )
        else:
          yield tfd.TruncatedNormal(
              0.0,
              jnp.ones_like(x),
              low=-2,
              high=2,
              name=f'initial_weight_matrix_{i}',
          )

    return lambda seed: tfd.JointDistributionCoroutine(
        _fn, use_vectorized_map=True, batch_ndims=0
    ).sample(seed=seed)

  prior = make_prior(**model_args)
  params = []
  losses = []
  for i in range(num_splits):
    if num_splits > 1:
      seed_i = jax.random.fold_in(seed, i)
    else:
      # Avoid changing the seed when num_splits is 1, for better comparisons
      # with previous experiments.
      seed_i = seed
    params_i, losses_i = ensemble_map(
        features,
        target,
        _neg_energy_fn,
        prior_d=prior,
        init_fn=_make_init_fn(prior),
        ensemble_size=((num_particles // num_splits) // jax.device_count()),
        learning_rate=learning_rate,
        num_epochs=num_epochs,
        seed=seed_i,
        batch_size=batch_size,
        prior_weight=prior_weight,
    )
    params.append(jax.tree_util.tree_map(np.array, params_i))
    losses.append(np.array(losses_i))
  params = jax.tree_util.tree_map(
      lambda *ts: np.concatenate(ts, axis=1), *params
  )
  losses = np.concatenate(losses, axis=1)
  return params, losses


def predict_bnf(
    features: ArrayT,
    observation_model: str,
    params: PyTree,
    model_args: dict[str, Any],
    quantiles: jax.Array,
    ensemble_dims: int = 2,
    approximate_quantiles: bool = False,
) -> tuple[jax.Array, list[jax.Array]]:
  """Predict new data from an existing BNF fit."""
  distribution = models.LikelihoodDist(observation_model)
  assert ensemble_dims >= 1

  forecast_inner = _make_forecast_inner(model_args, distribution)
  for _ in range(ensemble_dims - 1):
    forecast_inner = jax.vmap(forecast_inner, in_axes=(0, None))
  forecast_inner = jax.pmap(forecast_inner, in_axes=(0, None))
  axis = tuple(range(ensemble_dims))

  forecast_params = forecast_parameters_batched(
      features, params, distribution, forecast_inner
  )
  if distribution == models.LikelihoodDist.NORMAL:
    (means, scales) = forecast_params
    forecast_means = means
    forecast_quantiles = _get_percentile_normal(
        forecast_means,
        scales,
        quantiles,
        axis=axis,
        approximate=approximate_quantiles,
    )

  elif (
      distribution == models.LikelihoodDist.NB
      or distribution == models.LikelihoodDist.ZINB
  ):
    obs_d = _build_observation_distribution(distribution, forecast_params)
    forecast_means = obs_d.mean()
    forecast_quantiles = jax.vmap(
        lambda q: _get_nb_quantiles_root(obs_d, q, ensemble_axes=axis)
    )(jnp.array(quantiles))

  else:
    raise ValueError(f'Unknown distribution: {distribution}')

  return forecast_means, forecast_quantiles


def ensemble_map(
    features: ArrayT,
    target: ArrayT,
    neg_energy_fn: Callable[[PyTree, jax.Array, jax.Array], float],
    prior_d: tfd.Distribution,
    init_fn: Callable[[jax.Array], PyTree],
    ensemble_size: int,
    learning_rate: float,
    num_epochs: int,
    seed: jax.Array,
    batch_size: int | None = None,
    prior_weight: float = 1.0,
) -> tuple[PyTree, jax.Array]:
  """Fit an ensemble of MAP estimates.

  Args:
    features: Training features.
    target: Training targets.
    neg_energy_fn: The callable log-likelihood function.
    prior_d: `tfp.distributions.JointDistributionSequential` defining the prior.
    init_fn: Callable that, given a jax.random.PRNGKey, returns initial
      parameters for optimization.
    ensemble_size: `int` number of estimates to fit **per device**.
    learning_rate: `float` learning rate for optimization.
    num_epochs: `int` number of epochs of optimization to run.
    seed: `jax.random.PRNGKey` random seed.
    batch_size: `int` batch size to use during optimization.  The dataset is
      split into batches of this size along dimension 0.  If the batch size does
      not evenly divide the size of the dataset, the final ragged batch is
      skipped each epoch.  Default: None, indicating that the entire dataset
      should be used as a single batch.
    prior_weight: A positive `float`.  When computing the loss, the prior
      log-prob is multiplied by `prior_weight`.

  Returns:
    params: Inferred parameters.  The leftmost dimensions will be
      `(NUM_DEVICES, ensemble_size)`.
    losses: Array of shape (NUM_DEVICES, ensemble_size, num_epochs) -- the
      losses from each step (epoch) of optimization.
  """
  features = jnp.array(features)
  target = jnp.array(target)
  if batch_size is None:
    batch_size = target.shape[0]

  def _target_log_prob_fn(params, x_batch, y_batch):
    # given one pytree of model params and a batch of data,
    # return the loss on the batch.
    if prior_weight == 0.0:
      return -(
          neg_energy_fn(params, x_batch, y_batch)
          * (target.shape[0] / batch_size))
    else:
      return -(
          neg_energy_fn(params, x_batch, y_batch)
          * (target.shape[0] / batch_size)
          + prior_d.log_prob(params) * prior_weight)

  init_seed, opt_seed = jax.random.split(seed, 2)
  num_devices = jax.device_count()
  init_params = jax.vmap(jax.vmap(init_fn))(
      jax.random.split(init_seed, (num_devices, ensemble_size))
  )

  @jax.pmap
  @jax.vmap
  def _run(init_params, seed):
    optimizer = optax.adam(learning_rate)
    opt_state = optimizer.init(init_params)

    def _reshape_to_batches(t):
      t = jax.tree_util.tree_map(
          lambda v: v[: (v.shape[0] // batch_size) * batch_size], t
      )
      return jax.tree_util.tree_map(
          lambda v: v.reshape((-1, batch_size) + v.shape[1:]), t
      )

    def _one_epoch(carry, _):
      params, opt_state, seed = carry
      seed, permute_seed = jax.random.split(seed, 2)
      if batch_size < target.shape[0]:
        x, y = permute_dataset(features, target, permute_seed)
      else:
        x, y = features, target

      def _one_step(carry, batch):
        params, opt_state = carry
        batch_x, batch_y = batch
        loss, grads = jax.value_and_grad(_target_log_prob_fn)(
            params, batch_x, batch_y
        )
        updates, opt_state = optimizer.update(grads, opt_state)
        params = optax.apply_updates(params, updates)
        return (params, opt_state), loss

      (params, opt_state), losses = jax.lax.scan(
          _one_step,
          (params, opt_state),
          (_reshape_to_batches(x), _reshape_to_batches(y)),
      )
      return (params, opt_state, seed), losses.mean()

    (params, _, _), losses = jax.lax.scan(
        _one_epoch, (init_params, opt_state, seed), None, length=num_epochs
    )
    return params, losses

  return _run(
      init_params, jax.random.split(opt_seed, (num_devices, ensemble_size))
  )


def ensemble_vi(
    features: ArrayT,
    target: ArrayT,
    neg_energy_fn: Callable[[PyTree, jax.Array, jax.Array], float],
    prior_d: tfd.Distribution,
    ensemble_size: int,
    learning_rate: float,
    num_epochs: int,
    seed: jax.Array,
    sample_size: int = 10,
    num_samples: int = 30,
    kl_weight: float = 1.0,
    batch_size: int | None = None,
) -> tuple[
    tfp.distributions.JointDistributionSequential, jax.Array, jax.Array
]:
  """Fit an ensemble of surrogate posteriors.

  NOTE: For now, this function uses full-batch Variational Inference.

  Args:
    features: features for fitting.
    target: An array of target values.
    neg_energy_fn: The callable log-likelihood function.
    prior_d: `tfp.distributions.JointDistributionSequential` defining the prior.
    ensemble_size: `int` number of surrogate posteriors to fit **per device**.
    learning_rate: `float` learning rate for optimization.
    num_epochs: `int` number of epochs of optimization to run.
    seed: `jax.random.PRNGKey` random seed.
    sample_size: `int` number of Monte Carlo samples to use in estimating the
      variational divergence.
    num_samples: `int` number of posterior samples to return.
    kl_weight: A positive `float`.  During inference, we attempt to find a
      surrogate posterior `q(z)` that maximizes a version of the ELBO with the
      `KL(surrogate posterior || prior)` term scaled by `kl_weight` -- i.e.
      `E_z~q [log p(x|z)] - kl_weight * KL(q || p)`.
    batch_size: `int` batch size.  If specified, the log-prob in each step of VI
      is computed on a batch of this size, instead of on the whole dataset.
      Defaults to `None`.

  Returns:
    surrogate_posterior: A distribution over the (approximate) posterior.
      for the learned mean-field surrogate posteriors.  The leftmost
      dimensions will be (NUM_DEVICES, ensemble_size)
    losses: Array of shape (NUM_DEVICES, ensemble_size, num_epochs) -- the
      losses from each step of optimization.
    predictions: Array of num_samples samples of the BNF parameters from each of
    the (NUM_DEVICES, ensemble_size) surrogate posteriors.
  """
  features = jnp.array(features)
  target = jnp.array(target)

  if batch_size is not None:
    assert (
        target.shape[0] >= batch_size
    ), f'{batch_size=} exceeds {target.shape[0]=}'

  @functools.partial(jax.vmap, in_axes=(0, None, None))  # over MC samples
  @functools.partial(jax.vmap, in_axes=(0, None, None))  # over batch size
  def _target_log_prob_fn_inner(params, x_batch, y_batch):
    # NOTE: we divide by kl_weight here so that
    # tfp.vi.fit_surrogate_posterior_stateless minimizes:
    #     E_z~q [(log q(params) - _target_log_prob_fn(params)]
    #       = E_z~q [(log q(params) - log p(params))
    #                - neg_energy_fn(params, x, y) / kl_weight]
    # which is equivalent to minimizing:
    #     E_z~q [ kl_weight * (log q(params) - log p(params))
    #            - neg_energy_fn(params, x, y) ]
    return prior_d.log_prob(params) + (
        neg_energy_fn(params, x_batch, y_batch)
        * (target.shape[0] / y_batch.shape[0])
        / kl_weight
    )

  def _target_log_prob_fn(*params, seed):
    if batch_size is None:
      return _target_log_prob_fn_inner(params, features, target)
    batch = jax.random.permutation(
        seed, jnp.arange(target.shape[0]))[:batch_size]
    return _target_log_prob_fn_inner(params, features[batch], target[batch])

  def _make_surrogate_posterior(*params, batch_ndims=1):
    def _fn():
      for i in range(len(params) // 2):
        yield tfd.Normal(
            params[2 * i], 0.0001 + jax.nn.softplus(params[2 * i + 1])
        )

    return tfd.JointDistributionCoroutine(
        _fn, use_vectorized_map=True, batch_ndims=batch_ndims
    )

  init_seed, opt_seed = jax.random.split(seed, 2)
  num_devices = jax.device_count()
  init_params = make_vi_init(prior_d)(
      (num_devices, ensemble_size), seed=init_seed)

  @jax.pmap
  def _fit(init_params, fit_seed):
    surrogate_params, losses = tfp.vi.fit_surrogate_posterior_stateless(
        _target_log_prob_fn,
        _make_surrogate_posterior,
        init_params,
        optax.adam(learning_rate),
        num_epochs,
        sample_size=sample_size,
        jit_compile=False,
        seed=fit_seed,
    )
    return surrogate_params, losses

  @jax.pmap
  def _predict(surrogate_params, sample_seed):
    return _make_surrogate_posterior(*surrogate_params, batch_ndims=1).sample(
        num_samples, seed=sample_seed
    )

  fit_seed, sample_seed = jax.random.split(opt_seed, 2)
  surrogate_params, losses = _fit(
      init_params, jax.random.split(fit_seed, num_devices)
  )
  predictions = _predict(
      surrogate_params, jax.random.split(sample_seed, num_devices)
  )

  # Because we scaled the likelihood by `1 / kl_weight` in `_target_log_prob_fn
  # above (instead of scaling the KL(q||p) term) -- here we multiply the losses
  # by `kl_weight`.
  losses = np.transpose(losses, (0, 2, 1)) * kl_weight

  return (
      _make_surrogate_posterior(*surrogate_params, batch_ndims=2),
      losses,
      predictions,
  )
